<?php

namespace App\Controller;

use App\Exception;
use App\Templates;

use App\Exception\IntegrationException as IntegrationException;
use App\Exception\ParamsException as ParamsException;
use App\Exception\NetworkException as NetworkException;
use App\Exception\GeneralErrorException as GeneralErrorException;
use App\Exception\SqlException as SqlException;

use Psr\Http\Message\ResponseInterface as Response;
use Psr\Http\Message\ServerRequestInterface as Request;


class WebService {
    /** 
    *  Controller class pattern 
    *  - Class to handle request from Client
    *  - This class also take the responsibility to communicate with third-party services 
    *  and handle their responses
    */
    
    const API_URL             = "https://deliverydireto.com.br/admin-api/";
    const CLIENT_ID           = "5d574507-f55e-4da8-8a4f-6f174d394d86";
    const CLIENT_SECRET       = "-maWk_siYRKdktO5cyFS2IJWBpZD5avW2dKs";
    const TOKEN_REFRESH_TIME  = "21300";
    
    const FAILOVER_EMAIL = "delivery@teknisa.com";
    //const FAILOVER_EMAIL = "daniel.santos@teknisa.com";
    
    private $connection;
    private $integrationService;
    private $consumerService;
    private $branchService;
    private $dispatchService;
    private $service;
    private $util;
    private $authToken;

    public function __construct (\Doctrine\DBAL\Connection $connection,
                                    \App\Service\IntegrationService $integrationService,
                                    \App\Service\ConsumerService $consumerService,
                                    \App\Service\BranchService $branchService,
                                    \App\Service\DispatchService $dispatchService,
                                    \App\Service\IntegrationService $service,
                                    \App\Util\Util $util) {
            
        $this->connection         = $connection;
        $this->integrationService = $integrationService;
        $this->consumerService    = $consumerService;  
        $this->branchService      = $branchService;
        $this->dispatchService    = $dispatchService;
        $this->service            = $service;
        $this->util               = $util;
    }
    
    /**
     *  Function to get Orders and integrated from DeliveryDireto API 
     *  filtering orders by date (date avaiable on Reponse object paramter)
     *  
     *  @param Request $request
     *  @param Response $response
     *  @throws ParamsException Invalid Parameters on items/request
     *  @throws SqlException Query Exeception
     *  @throws Exception Generic Exception
     *  @throws NetworkException Error on communicate to another WebService
     *  @return Response Object
     * 
     */
    public function runWebService(Request $request, Response $response) {
        try {
            $req = $request->getParsedBody();
            
            if ($req["auth"] != "81743A8E8D41AF82A4807097CDF4BEC2DDDB9E6334BECCFED4D46FC783A8CA2E") 
                throw new \Exception("Auth inválido.", 666);
                
            $configuration   = $this->integrationService->getConfiguration();
            
            if(empty($configuration)){
                return $response->withStatus(200);
            }
            
            $ufTimezone      = (isset($configuration['IDUFTIMEZONE']) && !empty($configuration['IDUFTIMEZONE'])) ? $configuration['IDUFTIMEZONE'] : "SP";
            $trackingEnabled = (isset($configuration['IDRASTREIAENTREGA']) && ($configuration['IDRASTREIAENTREGA'] == "S"));
            
            date_default_timezone_set($this->util->getCurrentTimezone($ufTimezone));
            if($configuration['IDHABINTE'] == 'N') {
                return $response->withStatus(200);
            }

            $this->mailList = $this->util->getMailList();

            $configFile = file_get_contents(__DIR__ . "/../../configs/configs.json");
            $config     = json_decode($configFile, true);
            
            $branchs    = $this->branchService->getBranchs();

            if(empty($branchs)) {
                $this->service->disableIntegration();
                throw new ParamsException("Ocorreu um erro durante a autenticação com o Delivery Direto. Este erro desabilita a integração.
                    Não foi encontrado um token associado a uma unidade cadastrada.
                    Entre em contato com o Delivery Direto e solicite o token da unidade que deseja integrar os pedidos.
                    Após obter o token correto, faça a correção no Retail e habilite a integração novamente.
                    Para mais informações, consulte a seção 5d 'Configurando os Tokens das Unidades' no manual da integração.
                    Em caso de dúvidas, entre em contato com a equipe de suporte Teknisa.");
            }
            
            foreach($branchs as $branch) {
                
                if(empty($branch['CDFILIAL'])) {
                    $this->service->disableIntegration();
                    throw new ParamsException("O código da filial não foi encontrado. Este erro desabilita a integração.
                        Entre em contato com a equipe de suporte Teknisa.");
                }
                
                if(empty($branch['IDTOKENINTGERAL'])) {
                    $this->service->disableIntegration();
                    throw new ParamsException("O Token Geral não foi encontrado. Este erro desabilita a integração.
                        Acesse o Retail e insira o Token Geral. Em caso de dúvidas, entre em contato com a equipe de suporte Teknisa.");
                }
                
                if(empty($branch['IDSENHATOKENGERAL'])) {
                    $this->service->disableIntegration();
                    throw new ParamsException("A Senha Geral não foi encontrada. Este erro desabilita a integração.
                        Acesse o Retail e insira a Senha Geral. Em caso de dúvidas, entre em contato com a equipe de suporte Teknisa.");
                }
                
                $branch["IDTOKEN"] = trim($branch["IDTOKEN"]);
                
                $this->authUsername = trim($branch['IDTOKENINTGERAL']);
                $this->authPassword = trim($branch['IDSENHATOKENGERAL']);
                
                if(strlen($branch["IDTOKEN"]) != 36) {
                    $this->service->disableIntegration();
                    throw new ParamsException("Ocorreu um erro durante a autenticação com o Delivery Direto. Este erro desabilita a integração.
                        O token da unidade cadastrado não corresponde o padrão de tokens do Delivery Direto.
                        O token deve possuir 36 caracteres sem espaços. O token cadastrado contém " . strlen($branch["IDTOKEN"]) . " caracteres.
                        Entre em contato com o Delivery Direto e confirme se o token da unidade está correto.
                        Após obter o token correto, faça a correção no Retail e habilite a integração novamente.
                        Para mais informações, consulte a seção 5d 'Configurando os Tokens das Unidades' no manual da integração.
                        ATENÇÃO: Este erro está relacionado ao token da UNIDADE. Não está relacionado ao Token Geral.
                        Em caso de dúvidas, entre em contato com a equipe de suporte Teknisa.");
                }
                
                $authParams = [
                    "grant_type"    => "password",
                    "client_id"     => self::CLIENT_ID,
                    "client_secret" => self::CLIENT_SECRET,
                    "username"      => $this->authUsername,
                    "password"      => $this->authPassword
                ];
            
                //Access Token Validaton
                if(isset($branch['AUTHTKNDTCRIAC']) && !empty($branch['AUTHTKNDTCRIAC'])) {
                    if($branch['AUTHTKNDTCRIAC'] > time()) {
                        $this->authToken = $branch['AUTHTOKEN'];
                    } else {
                        $authTokenParam  = $this->getNewAuthToken($authParams, $branch["IDTOKEN"]);
                        $this->authToken = $authTokenParam['access_token'];
                    }
                } else {
                    $authTokenParam      = $this->getNewAuthToken($authParams, $branch["IDTOKEN"]);
                    $this->authToken     = $authTokenParam['access_token'];
                }
                
                $yesterday = date('Y-m-d', strtotime('-1 day'));

                // Check Raw Orders
                $rawOrdersParams = [
                    "showItems"             => "true",
                    "showExtras"            => "true",
                    "showMetadata"          => "true",
                    "includeCanceledOrders" => "true",
                    "dateStart"             => $yesterday
                ];

                $rawOrders   = json_decode($this->executeWebService("v1/orders", "GET", $rawOrdersParams, null, $branch["IDTOKEN"]), true);

                if(isset($rawOrdersParams)) {
                    unset($rawOrdersParams);
                }

                $this->service->validateCashierSinc($branch["CDFILIAL"]);
                // Process raw orders
                
                $orders = [];
                
                // -- START OF FILTERS --
                if(isset($rawOrders["data"]["orders"]) && !empty($rawOrders["data"]["orders"])) {
                    foreach($rawOrders["data"]["orders"] as $order) {
                        
                        //Filter Old Orders
                        if(empty($this->checkOlderOrders($order))) { 
                            continue;
                        }
                        
                        // Filter Canceled Orders
                        if($order["isCanceled"] == true || $order["status"] == "REJECTED") {
                            $this->cancelOrders($order["orderNumber"]);
                            continue;
                        }
                        
                        //Filter Orders With Other Status
                        if($order["status"] != "WAITING" && $order["status"] != "HIDDEN") {
                            continue;
                        }
                        
                        // Filter Same Orders Across Branchs
                        if($order["metadata"]["store"]["deliverydiretoId"] != $branch["IDTOKEN"]) {
                            $this->util->writeLog("execution", "", $dateLog. ": [Pedido Duplicado] #" . $order["orderNumber"] . " - Este pedido pertence a filial com DeliveryDiretoId " . $branch["IDTOKEN"] . " e, por isso, foi ignorado.\n");
                            continue;
                        }
                        
                        // Filter Duplicated Orders
                        if(!$this->service->validateDuplication($order["orderNumber"], $branch["CDFILIAL"])) {
                            continue;
                        }
                        array_push($orders, $order);
                    }
                }
                // -- END OF FILTERS --

                if(empty($orders)){
                    $this->dispatchPendingOrders();
                    continue;
                }
                
                foreach($orders as $order) {    
                    $this->util->writeLog('order', "$branch[CDFILIAL]-$order[id].json", $order);
                    
                    $order["branch"] = $branch["CDFILIAL"];
                    $order["idFrn"]  = $branch["IDTOKEN"];
                    
                    $dateLog = date("d/m/Y H:i:s");
                    
                    $this->util->writeLog("execution", "", $dateLog. ": [ Pedido Recebido] #" . $order["orderNumber"] . " - Filial: " . $order["branch"] . " - DeliveryDiretoId: " . $order["idFrn"] . "\n"); 

                    $this->connection->beginTransaction();
                    try {
                        $integratedOrder = $this->processOrder($order);
                        
                        $this->service->validateOrder($order["branch"], $integratedOrder["NRVENDAREST"]);
                        
                        $this->connection->commit();
                        
                        $orderStatusParams = [
                            "id" => $order["orderNumber"]
                        ];
                        
                        $orderStatusBody = [
                            "status" => "APPROVED"
                        ];
                        
                        $this->executeWebService("v1/orders/" . $order["orderNumber"] . "/status", "PUT", $orderStatusParams, $orderStatusBody, $branch["IDTOKEN"]);
                        
                        $dateLog = date("d/m/Y H:i:s");
                        $content = "[$dateLog - Pedido Processado]: #$order[orderNumber] - {Filial: $integratedOrder[CDFILIAL] / Número da venda: $integratedOrder[NRVENDAREST] / Número da comanda: $integratedOrder[NRCOMANDA] }\n";
                        
                        if($trackingEnabled) {
                            $orderType = !empty($order['type']) ? $order['type'] : "";
                            $this->dispatchService->registerPendingDispatchOrder($integratedOrder["CDFILIAL"], $integratedOrder['NRCOMANDA'], $integratedOrder['NRCOMANDAEXT'], $order["orderNumber"], $order["idFrn"], $orderType);
                        }
                        
                        $this->util->writeLog('order_processed', "$branch[CDFILIAL]-$order[id].json", $order);
                        $this->util->writeLog("execution", "", $dateLog. ": [Pedido Integrado] #" . $order["orderNumber"] . " -> Sucesso.\n");                    
                    } catch(NetworkException $e) {
                        $this->connection->rollback();
                        
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [  Erro de Rede  ] (NetworkException) -> {$e->getMessage()}\n");
                    } catch(ParamsException $e) {
                        $this->connection->rollback();
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [Erro Parametrizacao] " . "#$order[orderNumber] -> ". $e->getMessage() . "\n");
                        
                        if(!isset($branch["CDFILIAL"])) {
                            $branch["CDFILIAL"] = "undefined";
                        }
                        
                        $dateLog = date("d/m/Y H:i:s");
                        $content = "[ErrorLog - $dateLog]: Filial -> [$branch[CDFILIAL]]; orderNumber -> $order[orderNumber]; [Error -> {$e->getMessage()}]\n";
                        $this->util->writeLog('order_error', "$branch[CDFILIAL]-$order[id].json", $content);
                        $this->util->sendMail((!empty($this->mailList) ? $this->mailList : [self::FAILOVER_EMAIL]), [], 'Integração Teknisa x Delivery Direto', [
                          ":INTEGRACAO"    => "Delivery Direto",
                          ":NUMERO_PEDIDO" => $order["orderNumber"],
                          ":NOME_SERVIDOR" => $config["server"],
                          ":ERRO_MSG"      => $e->getMessage(),
                          ":JSON_PEDIDO"   => json_encode($order, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE),
                          ":GERADO_EM"     => date("d/m/Y H:i:s")
                        ]);
                        
                        // Cancelling Order
                        $orderStatusBody = [
                            "status"       => "HIDDEN",
                            "statusReason" => "Pedido cancelado."
                        ];

                        $this->executeWebService("v1/orders/" . $order["orderNumber"] . "/status", "PUT", null, $orderStatusBody, $order["idFrn"]);
                        
                    } catch(IntegrationException $e) {
                        $this->connection->rollback();
                        
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [Erro Integracao] " . "#$order[orderNumber] -> ". $e->getMessage() . "\n");
                        
                        if(!isset($branch["CDFILIAL"])) {
                            $branch["CDFILIAL"] = "undefined";
                        }
                        
                        $dateLog = date("d/m/Y H:i:s");
                        $content = "[ErrorLog - $dateLog]: Filial -> [$branch[CDFILIAL]]; orderNumber -> $order[orderNumber]; [Error -> {$e->getMessage()}]\n";
                        
                        $this->util->writeLog('order_error', "$branch[CDFILIAL]-$order[id].json", $content);
                        $this->util->sendMail((!empty($this->mailList) ? $this->mailList : [self::FAILOVER_EMAIL]), [], 'Integração Teknisa x Delivery Direto', [
                          ":INTEGRACAO"    => "Delivery Direto",
                          ":NUMERO_PEDIDO" => $order["orderNumber"],
                          ":NOME_SERVIDOR" => $config["server"],
                          ":ERRO_MSG"      => $e->getMessage(),
                          ":JSON_PEDIDO"   => json_encode($order, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE),
                          ":GERADO_EM"     => date("d/m/Y H:i:s")
                        ]);
                        
                        // Cancelling Order
                        $orderStatusBody = [
                            "status"       => "HIDDEN",
                            "statusReason" => "Pedido cancelado."
                        ];
                        
                        $this->executeWebService("v1/orders/" . $order["orderNumber"] . "/status", "PUT", null, $orderStatusBody, $order["idFrn"]);
                        
                    } catch(\Exception $e) {
                        $this->connection->rollback();
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [   Erro Geral   ] " . "#$order[orderNumber] -> " . $e->getMessage() . "\n");

                        if(!isset($branch["CDFILIAL"])) {
                            $branch["CDFILIAL"] = "undefined";
                        }
                        
                        $dateLog = date("d/m/Y H:i:s");
                        $content = "[ErrorLog - $dateLog]: Filial -> [$branch[CDFILIAL]]; orderNumber -> $order[orderNumber]; [Error -> {$e->getMessage()}]\n";
                        
                        $this->util->writeLog('order_error', "$branch[CDFILIAL]-$order[id].json", $content);
                        $this->util->sendMail((!empty($this->mailList) ? $this->mailList : [self::FAILOVER_EMAIL]), [], 'Integração Teknisa x Delivery Direto', [
                          ":INTEGRACAO"    => "Delivery Direto",
                          ":NUMERO_PEDIDO" => $order["orderNumber"],
                          ":NOME_SERVIDOR" => $config["server"],
                          ":ERRO_MSG"      => $e->getMessage(),
                          ":JSON_PEDIDO"   => json_encode($order, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE),
                          ":GERADO_EM"     => date("d/m/Y H:i:s")
                        ]);

                        // Cancelling Order
                        $orderStatusBody = [
                            "status"       => "HIDDEN",
                            "statusReason" => "Pedido cancelado."
                        ];

                        $this->executeWebService("v1/orders/" . $order["orderNumber"] . "/status", "PUT", null, $orderStatusBody, $order["idFrn"]);
                    }
                }
            }
            
            $this->itsAlive();
        } catch(NetworkException $e) {
            $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [  Erro de Rede  ] (NetworkException) -> {$e->getMessage()}\n");
        } catch(GeneralErrorException $e) {
            $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [   Erro Geral   ] (GeneralErrorException) -> {$e->getMessage()}\n");
            
            $this->service->disableIntegration();
            
            $template = file_get_contents(__DIR__ ."/../../templates/general-error.html");
            
            $this->util->sendMail((!empty($this->mailList) ? $this->mailList : [self::FAILOVER_EMAIL]), [], "Integração Teknisa x Delivery Direto", $template, [
                ":INTEGRACAO"    => "Delivery Direto",
                ":NOME_SERVIDOR" => $config["server"],
                ":ERRO_MSG"      => $e->getMessage(),
                ":GERADO_EM"     => date("d/m/Y H:i:s")
            ]);
            
        } catch(\Exception $e) {
            $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [Erro Desconhecido] -> {$e->getMessage()}\n");
        }

        return $response->withStatus(200);
    }
    
    /**
     *  Function to integrate a order into the Teknisa systems
     *   
     *  @param  array with DeliveryDireto Order information
     *  @return array with new codes generted for the order in Teknisa
     *          system
     */
    private function processOrder($order) {
        $generalParams = $this->service->generalParameters($order["branch"]);

        $generalParams["CPF"] = ""; // CPF para impressão na nota
        
        if(isset($order['registeredDocument']) && !empty($order['registeredDocument'])) {
            $generalParams["CPF"] = preg_replace( "/[^0-9]/", "", $order['registeredDocument']);
        }
        else if (isset($order["customer"]["document"]) || !empty($order["customer"]["document"])) {
            $generalParams["CPF"] = preg_replace( "/[^0-9]/", "", $order["customer"]["document"]);
        }

        // TOGO - Insert branch address to order
        if($order["type"] == "TAKEOUT") {
            $branchInfoAddress                   = $this->branchService->getBranchAddress($order["branch"]);
            $order["address"]["state"]           = $branchInfoAddress["SGESTADO"];
            $order["address"]["city"]            = $branchInfoAddress["CDMUNICIPIO"];
            $order["address"]["neighborhood"]    = $branchInfoAddress["NMBAIRFILI"];
            $order["address"]["number"]          = preg_replace("/[^0-9]/", "", $branchInfoAddress["DSENDEFILI"]);
            $order["address"]["street"]          = str_replace($order["address"]["number"], "", $branchInfoAddress["DSENDEFILI"]);
            $order["address"]["zipcode"]         = $branchInfoAddress["NRCEPFILI"];
            $order["address"]["complement"]      = "";
            $order["address"]["reference_point"] = "";
            $order["address"]["lat"]             = "";
            $order["address"]["lng"]             = "";
        }
        
        $consumer = $this->consumerService->processConsumer($generalParams, $order);
        return $this->service->openOrderBill($generalParams, $order, $consumer, $order["items"]);
    }
    
    /**
     *  Function to integrate a order into the Teknisa systems
     *   
     *  @param  array with DeliveryDireto Order information
     *  @return array with new codes generted for the order in Teknisa
     *          system
     */
    private function cancelOrders($orderNumber) {
        $this->service->cancelOrder($orderNumber);
    }
    
    private function checkOlderOrders($order) {
        
        $utcTimezone = new \DateTimeZone("UTC");
        
        $now = new \DateTime("now", $utcTimezone);
        $now->modify("-55 minutes");
        
        $orderCreated = new \DateTime($order["created"], $utcTimezone);
        
        if($orderCreated > $now) {
            return $order;
        } else {
            return 0;
        }
    }
    
    /* 
     * Controle de status do delivery direto 
    */
    private function dispatchPendingOrders() {
        $pendingOrdersFile = $this->util->getPendingOrdersFile();
        
        $this->filterDispatchFile($pendingOrdersFile);
        $this->processPendingDispatchOrders($pendingOrdersFile);
        $this->processPendingDeliveredOrders($pendingOrdersFile);
        
        return 0;
    }
    
    private function filterDispatchFile($pendingOrdersFile) {
        $pendingDispatchFile = file_get_contents($pendingOrdersFile);   
        $pendingDispatchData = json_decode($pendingDispatchFile, true);

        if(empty($pendingDispatchData)) {
            return 0;
        }

        $activeList = [];
        foreach ($pendingDispatchData as $key => $currentOrder) {
            
            $diff = (time() - $currentOrder['createdAt']);
            if($diff > 28800) { // 8h
                $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [ Pedido Delivery] #" . $currentOrder['nrComandaExt'] . " expirou (nao foi expedido dentro do tempo limite de 8h)." . PHP_EOL);
                continue;
            }

            $activeList[$key] = $currentOrder;
        }

        file_put_contents($pendingOrdersFile, json_encode($activeList, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE));
    }

    private function processPendingDispatchOrders($pendingOrdersFile) {
        $pendingDispatchFile = file_get_contents($pendingOrdersFile);   
        $pendingDispatchData = json_decode($pendingDispatchFile, true);

        if(empty($pendingDispatchData))
            return 0;

        $cdFilialArr     = [];
        $nrComandaArr    = [];
        $nrComandaExtArr = [];
        
        foreach($pendingDispatchData as $pendingDispatchOrder) {
            
            if(!empty($pendingDispatchOrder['dispatchedAt']) || $pendingDispatchOrder['delivered']) {
                continue;
            }
            
            if(!isset($pendingDispatchOrder['cdFilial']) || empty($pendingDispatchOrder['cdFilial'])) {
                continue;                
            }            
            
            if(!isset($pendingDispatchOrder['nrComanda']) || empty($pendingDispatchOrder['nrComanda'])) {
                continue;                
            }            
            
            if(!isset($pendingDispatchOrder['nrComandaExt']) || empty($pendingDispatchOrder['nrComandaExt'])) {
                continue;                
            }            

            array_push($cdFilialArr,     $pendingDispatchOrder['cdFilial']);
            array_push($nrComandaArr,    $pendingDispatchOrder['nrComanda']);
            array_push($nrComandaExtArr, $pendingDispatchOrder['nrComandaExt']);
        }
        
        if(empty($cdFilialArr)) 
            return 0;
        
        if(empty($nrComandaArr)) 
            return 0;
        
        if(empty($nrComandaExtArr)) 
            return 0;
        
        $orders = $this->dispatchService->getPendingDispatchOrders($cdFilialArr, $nrComandaArr, $nrComandaExtArr);

        if(empty($orders)) {
            return 0;
        }
        
        foreach($orders as $order) {
            $key = "{$order['CDFILIAL']}|{$order['NRCOMANDA']}|{$order['NRCOMANDAEXT']}";
            
            $idToken = $pendingDispatchData[$key]['idToken'];
            
            if($order['IDSTCOMANDA'] == '4') {
                if(isset($pendingDispatchData[$key])) {
                    $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [Pedido Cancelado] #" . $pendingDispatchData[$key]['nrComandaExt'] . "- Removido da fila de controle de entregador (motivo: pedido cancelado)." . PHP_EOL);
                    
                    // Cancelling Order
                    $orderStatusBody = [
                        "status"       => "HIDDEN",
                         "statusReason" => "Pedido cancelado."
                     ];
                    
                    $this->executeWebService("v1/orders/" . $order["NRCOMANDAEXT"] . "/status", "PUT", null, $orderStatusBody, $idToken);
                    $pendingDispatchData[$key]['canceled'] = true;
                }
            }
            
            if($order['IDSTCOMANDA'] == '5' || $order['IDSTCOMANDA'] == 'X') {
                if(isset($pendingDispatchData[$key])) {
                    $reference = $pendingDispatchData[$key]['reference'];
                    $orderType = $pendingDispatchData[$key]['orderType'];

                    // Dispatching Order
                    $orderStatusBody = [
                        "status"       => "IN_TRANSIT",
                    ];
                    
                    $this->executeWebService("v1/orders/" . $order["NRCOMANDAEXT"] . "/status", "PUT", null, $orderStatusBody, $idToken);
                    
                    if ($orderType == "DELIVERY") {
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [ Pedido Delivery] #" . $pendingDispatchData[$key]['nrComandaExt'] . " - Pedido saiu para entrega." . PHP_EOL);
                        $pendingDispatchData[$key]['dispatchedAt'] = time();   
                    } elseif ($orderType == "TAKEOUT") { 
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [ Pedido Retirada] #" . $pendingDispatchData[$key]['nrComandaExt'] . " - Pedido pronto para retirada." . PHP_EOL);
                        $pendingDispatchData[$key]['dispatchedAt'] = time();
                    } elseif ($orderType == "TABLE") { 
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [   Pedido Mesa  ] #" . $pendingDispatchData[$key]['nrComandaExt'] . " - Pedido pronto para retirada." . PHP_EOL);
                        $pendingDispatchData[$key]['dispatchedAt'] = time();
                    } else {
                        $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . " [ Pedido Removido ] #" . $pendingDispatchData[$key]['nrComandaExt'] . " - Pedido removido da fila de entrega." . PHP_EOL);
                        $pendingDispatchData[$key]['canceled'] = true;
                    }
                }
            }
        }
        
        $activeList = [];
        
        foreach ($pendingDispatchData as $key => $currentOrder) {
            if($currentOrder['canceled']) {
                continue;
            }
            $activeList[$key] = $currentOrder;
        }
        
        file_put_contents($pendingOrdersFile, json_encode($activeList, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE));
    }

    private function processPendingDeliveredOrders($pendingOrdersFile) {
        $pendingDeliveredFile = file_get_contents($pendingOrdersFile);   
        $pendingDeliveredData = json_decode($pendingDeliveredFile, true);
        
        if(empty($pendingDeliveredData))
            return 0;
            
        
        foreach($pendingDeliveredData as $key => $pendingDeliveredOrder) {
            
            if(empty($pendingDeliveredOrder['dispatchedAt'])) {
                continue;
            }

            $diff = (time() - $pendingDeliveredOrder['dispatchedAt']);
            if($diff >= 3600) {//3600
                $reference = $pendingDeliveredOrder['reference'];
                
                // Dispatching Order
                $orderStatusBody = [
                    "status"       => "DONE",
                ];
                
                $this->executeWebService("v1/orders/" . $pendingDeliveredOrder["nrComandaExt"] . "/status", "PUT", null, $orderStatusBody, $pendingDeliveredOrder['idToken']);
                
                $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [ Pedido Delivery] #" . $pendingDeliveredOrder["nrComandaExt"] . " - Pedido entregue." . PHP_EOL);
                
                $pendingDeliveredData[$key]['delivered'] = true;
            }
        }
    
        $newPendingList = [];
        
        foreach($pendingDeliveredData as $key => $pendingDeliveredOrder) {
            if($pendingDeliveredOrder['delivered']) {
                continue;
            }
            $newPendingList[$key] = $pendingDeliveredOrder;
        }

        file_put_contents($pendingOrdersFile, json_encode($newPendingList, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE));
 
    }
	
    /**
     *  Function to send a request for exeternal service
     * 
     *  @param $function - name of service to be called
     *  @param $mehtod - type method of request
     *  @custom
     *  @return array with ids of valiz
     */
    private function executeWebService ($function, $method, $params=[], $body=[], $idToken) {

        $headers = array(
            "X-DELIVERYDIRETO-ID: $idToken",
            "X-DELIVERYDIRETO-CLIENT-ID: " . self::CLIENT_ID,
            "Content-Type: application/json"
        );
        
        if(isset($this->authToken) && !empty($this->authToken)) {
            array_push($headers, 'Authorization: Bearer ' . $this->authToken);
        }

        $service_url = self::API_URL.$function;
        
        if(isset($params) && !empty($params)) {
            
            $string = '?';
            foreach ($params as $key => $value) {
                $string .= $key . '=' . urlencode($value) . '&';
            }
            $string = rtrim($string, '&');
            $service_url = $service_url . $string;
        }

        $curl = curl_init();
        
        curl_setopt($curl, CURLOPT_URL, $service_url);
        curl_setopt($curl, CURLOPT_CUSTOMREQUEST, $method);
        curl_setopt($curl, CURLOPT_POSTFIELDS, json_encode($body));
        curl_setopt($curl, CURLOPT_HTTPHEADER, $headers);
        curl_setopt($curl, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
        curl_setopt($curl, CURLOPT_RETURNTRANSFER, true);
        curl_setopt($curl, CURLOPT_TIMEOUT, 25);
        curl_setopt($curl, CURLOPT_FAILONERROR, true);
        curl_setopt($curl, CURLOPT_SSL_VERIFYPEER, false);
        curl_setopt($curl, CURLINFO_HEADER_OUT, true);
        
        $curl_response = curl_exec($curl);
        $httpCode      = curl_getinfo($curl, CURLINFO_HTTP_CODE);
        
        if (!empty(curl_error($curl)) || stripos($curl_response, "html>") != false) {
            $errorMessage = date("d/m/Y H:i:s") . " - " . "cURL error: " . curl_error($curl);
            $this->util->writeLog('network', date('H-i-s') . '.txt', $this->util->formatNetworkError([
                'eventTime' => date("d/m/Y H:i:s"),
                'header'    => $headers,
                'url'       => $service_url,
                'method'    => $method,
                'params'    => $params,
                'body'      => $body,
                'httpCode'  => $httpCode,
                'error'     => curl_error($curl),
                'response'  => $curl_response
            ]));
        
            // if($httpCode == 401) {
            //     $this->service->disableIntegration();
            //     throw new ParamsException("Ocorreu um erro durante a autenticação com o Delivery Direto. Este erro desabilita a integração.
            //         As credenciais de acesso (Token Geral, Senha Geral e X-DELIVERYDIRETO-ID) foram recusadas.
            //         Entre em contato com o Delivery Direto e valide as credenciais de acesso.
            //         Após validar os tokens, faça as devidas correções no Retail e habilite a integração novamente.
            //         Em caso de dúvidas, entre em contato com a equipe de suporte Teknisa.");
            // }
            
            if($httpCode == 404 && $body['status'] == "IN_TRANSIT") {
                $this->util->writeLog("execution", "", date("d/m/Y H:i:s") .
                    ": [Falha na Atualizacao do Pedido] Ocorreu um erro ao atualizar o status do pedido " . substr($function, -15, 8) . ".
                     O pedido nao foi localizado no Delivery Direto.\n");
            }
            else if(substr(strval($httpCode), 0, 1) == "5") {
                throw new NetworkException("Ocorreu um erro durante a comunicação da integração com o Delivery Direto.
                    A causa mais provável é uma indisponibilidade do servidor do Delivery Direto.
                    Verifique com o Delivery Direto se houve alguma interrupção no serviço.");
            } else {
                $this->util->writeLog("execution", "", date("d/m/Y H:i:s") ." [Erro de Requisicao] Ocorreu um erro durante uma requisicao com o Delivery Direto.
                     Este erro nao impedira o funcionamento da integracao.
                     Caso este erro se repita, entre em contato com o suporte da Teknisa.\n");
            }
        }
        
        curl_close($curl);
        return $curl_response;
    }
    
    public function showVersion(Request $request, Response $response) {
        $sql        = "SELECT CDFILIAL FROM FILIAL";
        $result     = $this->connection->fetchAll($sql);

        $versionInfo = $this->getVersionInfo();

        return $response->withJson([
            'version'     => $versionInfo['version'],
            'revision'    => $versionInfo['revision'],
            'date'        => $versionInfo['date'],
            'branchs'     => count($result),
        ]);
    }
    
    private function getVersionInfo() {
        $file = file_get_contents(__DIR__ . "/../../version.json");
        return json_decode($file, true);
    }

    public function itsAlive() {
        $executionLogFilePath  = $this->util->getExecutionLog();
        $executionLogTimeStamp = filemtime($executionLogFilePath);
        if((time() - $executionLogTimeStamp) >= 3600) {
            $this->util->writeLog("execution", "", date("d/m/Y H:i:s") . ": [   Integridade  ] Verificacao de integridade.\n");
        }
    }
    
    public function getNewAuthToken($authParams, $idToken) {
        $this->util->writeLog("execution", "", date("d/m/Y H:i:s").": [  Autenticacao  ] Realizando nova autenticacao: ");

        $body = [
            "grant_type"    => "password",
            "client_id"     => $authParams['client_id'],
            "client_secret" => $authParams['client_secret'],
            "username"      => $authParams['username'],
            "password"      => $authParams['password']
        ];

        $response = json_decode($this->executeWebService("token", "POST", null, $body, $idToken), true);

        $fileName = date("d") . ".txt";
        
        if (empty($response)) {
            $this->util->writeLog(
                'network', 
                $fileName, 
                $this->getContentAuthLog(['Content-Type: application/json'], $targetUrl, $params, $httpCode, curl_error($curl), $curl_response)
            );
            
            curl_close($curl);
            throw new \Exception("Não foi possível recuperar o token de acesso à API. Sem resposta do endpoint.");
        } 
        
        if(!isset($response["access_token"])) {
            $this->util->writeLog("execution", "", "Erro ao obter novo token\n");

            $this->util->writeLog(
                'network', 
                $fileName, 
                $this->getContentAuthLog(['Content-Type: application/json'], $targetUrl, $params, $httpCode, curl_error($curl), $curl_response)
            );
            
            curl_close($curl);
            throw new \Exception("Não foi possível recuperar o token de acesso à API. Erro de credencial.");        
        }
        
        $this->util->writeLog("execution", "", "Sucesso\n");        
        $this->setAccessToken($response, $idToken);
        
        return $response;
    }
    
    public function setAccessToken ($response, $idToken) {
        $sql = "
            UPDATE INTEGRACAOFILI
               SET AUTHTOKEN      = :AUTHTOKEN,
                   AUTHTKNDTCRIAC = :AUTHTKNDTCRIAC
             WHERE IDTOKEN        = :IDTOKEN
        ";
        
        $authtkndtcriac = time() + self::TOKEN_REFRESH_TIME;
    	
    	$params = [
    	    ":AUTHTOKEN"      => $response['access_token'],
            ":AUTHTKNDTCRIAC" => $authtkndtcriac,
            ":IDTOKEN"        => $idToken
    	];
    	
    	$this->connection->executeQuery($sql, $params);
    }
    
    public function debugOrders(Request $request, Response $response)
    {
        $req = $request->getParsedBody();

        if ($req["auth"] != "81743A8E8D41AF82A4807097CDF4BEC2DDDB9E6334BECCFED4D46FC783A8CA2E") 
            throw new \Exception("Auth inválido.", 666);
            
        $configuration   = $this->integrationService->getConfiguration();
        $ufTimezone      = (isset($configuration['IDUFTIMEZONE']) && !empty($configuration['IDUFTIMEZONE'])) ? $configuration['IDUFTIMEZONE'] : "SP";
        $trackingEnabled = (isset($configuration['IDRASTREIAENTREGA']) && ($configuration['IDRASTREIAENTREGA'] == "S"));
        
        date_default_timezone_set($this->util->getCurrentTimezone($ufTimezone));
        
        if($configuration['IDHABINTE'] == 'N') {
            return $response->withStatus(200);
        }

        $configFile = file_get_contents(__DIR__ . "/../../configs/configs.json");
        $config     = json_decode($configFile, true);
        
        $date       = gmdate('Y-m-d H:i:s', time() - (5*60));
        
        $order = $req["data"];
        
        $cdFilial = "0001";
        $idFrn    = "mhcukDOBalvTjcW1H1uJo5w4qApKdrSTfFYL";
        
        // Descomente para validar horário do pedido.
        // $newOrder = $this->checkOlderOrders($order);
        // if(!empty($newOrder)) { 
        //     echo("Pedido novo");
        //     //array_push($newOrders, $newOrder);
        // } else {
        //     echo("Pedido velho");
        // }
        
        //Início das validações
        
        //Filter Old Orders
        if(empty($this->checkOlderOrders($order))) {  
            echo ("Não passou pela validação de pedidos antigos.");
            return 0;
        }
        
        // Filter Canceled Orders
        if($order["isCanceled"] == true || $order["status"] == "REJECTED") {
            $this->cancelOrders($order["orderNumber"]);
            echo ("Não passou pela validação de pedidos cancelados.");
            return 0;
        }
        
        //Filter Orders With Other Status
        if($order["status"] != "WAITING" && $order["status"] != "HIDDEN") {
            echo ("Pedido com status diferente de WAITING e HIDDEN.");
            return 0;
        }
        
        // Filter Same Orders Across Branchs
        // if($order["metadata"]["store"]["deliverydiretoId"] != $idFrn) {
        //     echo ("Não passou pela validação de pedidos repetidos entre as filiais.");
        //     return 0;
        // }
        
        // Filter Duplicated Orders
        if(!$this->service->validateDuplication($order["orderNumber"], $cdFilial)) {
            echo ("Não passou pela validação de pedidos duplicados.");
            return 0;
        }
        
        // Fim das validações
        
        $this->util->writeLog('order', "0001-$order[id].json", $order);
        
        $order["branch"] = "0001";
        $order["idFrn"]  = "mhcukDOBalvTjcW1H1uJo5w4qApKdrSTfFYL";
        
        $dateLog = date("d/m/Y H:i:s");
        $this->util->writeLog("execution", "", $dateLog. ": [ Pedido Recebido] #" . $order["orderNumber"] . " - Filial: " . $order["branch"] . " - DeliveryDiretoId: " . $order["idFrn"] . "\n"); 

        $this->connection->beginTransaction();
        try {
            $integratedOrder = $this->processOrder($order);
            
            $this->service->validateOrder($order["branch"], $integratedOrder["NRVENDAREST"]);
            
            $this->connection->commit();

            $dateLog = date("d/m/Y H:i:s");
            $content = "[$dateLog - Pedido Processado]: #$order[orderNumber] - {Filial: $integratedOrder[CDFILIAL] / Número da venda: $integratedOrder[NRVENDAREST] / Número da comanda: $integratedOrder[NRCOMANDA] }\n";
            
            // if($order["type"] == "DELIVERY") {
            //     $this->dispatchService->registerPendingDispatchOrder($integratedOrder["CDFILIAL"], $integratedOrder['NRCOMANDA'], $integratedOrder['NRCOMANDAEXT'], $order["orderNumber"], $order["idFrn"]);
            // } else {
            //     $this->dispatchService->registerTogoOrder($integratedOrder["CDFILIAL"], $integratedOrder['NRCOMANDA'], $integratedOrder['NRCOMANDAEXT'], $order["orderNumber"], $order["idFrn"], $integratedOrder["IMPAUTO"]);
            // }
            
            $this->util->writeLog('order_processed', "$order[branch]-$order[id].json", $order);
            $this->util->writeLog("execution", "", $dateLog. ": [Pedido Integrado] #" . $order["orderNumber"] . " -> Sucesso.\n");                    
            
            return $response->withJson([
                'success' => true,
                'orderId' => $order['orderNumber']
            ]);
        }catch(ParamsException $e) {
            $this->connection->rollback();
            echo($e->getMessage() . "\n");

        } catch(IntegrationException $e) {
            $this->connection->rollback();
            echo($e->getMessage() . "\n");

        } catch(\Exception $e) {
            $this->connection->rollback();
            echo($e->getMessage() . "\n");
        }
    }
}
